from __future__ import annotations

from typing import List, Dict, Tuple

from aasm.modules.instruction import Instruction
from aasm.modules.type import Type
from aasm.utils.exception import PanicException


class Target:
    def __init__(self, target: str):
        self.name = target

    def __repr__(self):
        return f"Target[{self.name}]"

    def __str__(self):
        return f"Target[{self.name}]"


class Module:
    def __init__(self, module_code_lines: List[str]):
        self.name = None
        # TODO: Change targets and types into classes
        self.targets: List[Target] = []
        self.types: List[Type] = []
        self.instructions: List[Instruction] = []
        self.preambles: Dict[str, List[str]] = {}
        self.impls: Dict[Tuple[str, str], List[str]] = {}
        self.description: List[str] = []

        self._in_targets = False
        self._in_instructions = False
        self._in_preamble = False
        self._in_impl = False
        self._in_init = False
        self._in_types = False
        self._in_description = False
        self._current_type = None
        self._current_target = None
        self._current_instruction = None

        self._parse_module_code(module_code_lines)
        # TODO: validate module -- check that all instructions are implemented for all targets, has a name etc.
        self._validate_module()

    def _validate_module(self):
        # has name
        if self.name is None:
            raise PanicException(
                "Error while loading unkown module",
                "Module has no name",
                "Please add a name to the module using !name",
            )
        # has at least one target
        if len(self.targets) == 0:
            raise PanicException(
                f"Error while loading module {self.name}",
                "Module has no targets",
                "Please add at least one target to the module using !targets",
            )

        # each type has an init for each target
        for type in self.types:
            for target in self.targets:
                if target.name not in type.init_lines:
                    raise PanicException(
                        f"Error while loading module {self.name}",
                        f"Type {type.name} has no init for target {target.name}",
                        "Please add an init for the target using !init [type] [target]",
                    )

        # each instruction has impl for each target
        for target in self.targets:
            for instruction in self.instructions:
                try:
                    self.impls[(target.name, instruction.opcode)]
                except KeyError:
                    raise PanicException(
                        f"Error while loading module {self.name}",
                        f"Instruction {instruction.opcode} has no impl for target {target.name}",
                        "Please add an impl for the target using !impl [instruction] [target]",
                    )

    def get_args_for_instruction(self, instruction_name: str) -> Dict[str, List[Type]]:
        for instruction in self.instructions:
            if instruction.opcode == instruction_name:
                return instruction.args_dict
        return {}

    def does_target(self, target: str) -> bool:
        return target in [target.name for target in self.targets]

    def _reset_scope(self):
        self._in_targets = False
        self._in_instructions = False
        self._in_preamble = False
        self._in_impl = False
        self._in_init = False
        self._in_types = False
        self._in_description = False
        self._current_target = None
        self._current_type = None
        self._current_instruction = None

    def _parse_module_code(self, lines: List[str]):
        for line in lines:
            tokens = line.strip().split()
            tokens = [token.strip().strip(",") for token in tokens]
            match tokens:
                case ["!name", name]:
                    self._reset_scope()
                    self.name = name
                case ["!types"]:
                    self._reset_scope()
                    self._in_types = True
                case ["!description"]:
                    self._reset_scope()
                    self._in_description = True
                case ["!targets"]:
                    self._reset_scope()
                    self._in_targets = True
                case ["!instructions"]:
                    self._reset_scope()
                    self._in_instructions = True
                case ["!preamble", target]:
                    self._reset_scope()
                    self._in_preamble = True
                    self._current_target = target
                case ["!impl", instruction, target]:
                    self._reset_scope()
                    self._in_impl = True
                    self._current_target = target
                    self._current_instruction = instruction.upper()
                case ["!init", type, target]:
                    self._reset_scope()
                    self._in_init = True
                    self._current_target = target
                    self._current_type = type
                case _:
                    if len(tokens) == 0:
                        continue
                    elif tokens[0].startswith("#"):
                        continue
                    elif tokens[0].startswith("!"):
                        raise PanicException(
                            "Invalid line: " + line,
                            "Unkown module directive",
                            "Only module directives can start with !",
                        )
                    elif self._in_targets:
                        if len(tokens) != 1:
                            raise PanicException(
                                "Invalid target line: " + line,
                                "Multiple tokens in target line",
                                "Target lines must have exactly one token: e.g. spade",
                            )
                        self.targets.append(Target(tokens[0]))
                    elif self._in_instructions:
                        if self.name is None:
                            raise PanicException(
                                "Invalid instruction line: " + line,
                                "Module name is undefined",
                                "Module name must be defined before instructions. Define module name with !name [name]",
                            )
                        else:
                            self.instructions.append(
                                Instruction(
                                    self.name, self.types, tokens[0], tokens[1:]
                                )
                            )
                    elif self._in_preamble:
                        if self._current_target is None:
                            raise PanicException(
                                "Invalid preamble line: Target is undefined: " + line,
                                "Target is undefined",
                                "Target must be defined before preamble. Define target with !preamble [target]",
                            )
                        self.preambles.setdefault(self._current_target, []).append(line)
                    elif self._in_impl:
                        if self._current_target is None:
                            raise PanicException(
                                "Invalid impl line: Target is undefined: " + line,
                                "Target is undefined",
                                "Target must be defined before impl. Define target with !impl [instruction] [target]",
                            )
                        if self._current_instruction is None:
                            raise PanicException(
                                "Invalid impl line: Instruction is undefined: " + line,
                                "Instruction is undefined",
                                "Instruction must be defined before impl. Define instruction with !impl [instruction] [target]",
                            )
                        self.impls.setdefault(
                            (self._current_target, self._current_instruction.upper()),
                            [],
                        ).append(line)
                    elif self._in_types:
                        if len(tokens) != 1:
                            raise PanicException(
                                "Invalid type line: " + line,
                                "Multiple tokens in type line",
                                "Type lines must have exactly one token: e.g. int64",
                            )
                        if self.name is None:
                            raise PanicException(
                                "Invalid instruction line: " + line,
                                "Module name is undefined",
                                "Module name must be defined before instructions. Define module name with !name [name]",
                            )
                        else:
                            self.types.append(Type(tokens[0], self.name))
                    elif self._in_init:
                        if self._current_target is None:
                            raise PanicException(
                                "Invalid init line: Target is undefined: " + line,
                                "Target is undefined",
                                "Target must be defined before init. Define target with !init [type] [target]",
                            )
                        if self._current_type is None:
                            raise PanicException(
                                "Invalid init line: Type is undefined: " + line,
                                "Type is undefined",
                                "Type must be defined before init. Define type with !init [type] [target]",
                            )
                        # find the type
                        found = False
                        for t in self.types:
                            if t.name == self._current_type:
                                t.add_init_line(self._current_target, line)
                                found = True
                                break
                        if not found:
                            raise PanicException(
                                "Invalid init line: Type is undefined: " + line,
                                f"Unkown type {self._current_type}",
                                "Type must be defined before init. Define type with !init [type] [target]",
                            )

                    elif self._in_description:
                        self.description.append(line)
                    else:
                        raise PanicException(
                            "Invalid line: " + line,
                            "Unkown line",
                            "Line is not a module directive, target, instruction, preamble or impl",
                        )

    def __repr__(self):
        return (
            f"Module[{self.name}] ("
            + repr(self.targets)
            + "\n"
            + repr(self.types)
            + "\n"
            + repr(self.instructions)
            + "\n"
            + repr(self.preambles)
            + "\n"
            + repr(self.impls)
            + ")"
        )

    def __str__(self):
        return (
            f"Module[{self.name}] (\n"
            + str(self.description)
            + "\n"
            + str(self.targets)
            + "\n"
            + str(self.types)
            + "\n"
            + str(self.instructions)
            + "\n"
            + str(self.preambles)
            + "\n"
            + str(self.impls)
            + ")"
        )



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.intermediate.argument import Argument
from aasm.intermediate.instruction import (
    LogsCritical,
    LogsDebug,
    LogsError,
    LogsInfo,
    LogsWarning,
)
from aasm.utils.validation import print_valid_logs_levels

if TYPE_CHECKING:
    from parsing.state import State


def op_LOGS(state: State, level: str, args: List[str]) -> None:
    state.require(
        state.in_action, "Not inside any action", f"LOGS can be used inside actions."
    )

    intermediate_args: List[Argument] = []
    for arg in [Argument(state, arg) for arg in args]:
        intermediate_args.extend(arg.create_all_possible_types(state))

    match level:
        case "debug":
            state.last_action.add_instruction(LogsDebug(intermediate_args))

        case "info":
            state.last_action.add_instruction(LogsInfo(intermediate_args))

        case "warning":
            state.last_action.add_instruction(LogsWarning(intermediate_args))

        case "error":
            state.last_action.add_instruction(LogsError(intermediate_args))

        case "critical":
            state.last_action.add_instruction(LogsCritical(intermediate_args))

        case _:
            state.panic(
                f"Unexpected error: LOGS {level} {args}",
                f"Supported LOGS levels are: {print_valid_logs_levels()}",
            )



"""Agents Assembly translator"""

__version__ = "0.1.5"

from aasm.generating.code import Code
from aasm.modules.module import Module
from aasm.generating.python_spade import get_spade_code
from aasm.generating.python_module import get_modules_for_target
from aasm.utils.exception import PanicException



from aasm.intermediate.argument import ModuleVariable


class Type:
    def __init__(self, name: str, module: str):
        self.name = name
        self.type_class = type(name, (ModuleVariable,), {})
        self.module = module
        self.init_lines = {}

    def full_qualified_name(self) -> str:
        return f"{self.module}::{self.name}"

    def add_init_line(self, target: str, line: str):
        if target in self.init_lines:
            self.init_lines[target].append(line)
        else:
            self.init_lines[target] = [line]

    def __eq__(self, other):
        if isinstance(other, Type):
            return self.name == other.name and self.module == other.module
        return False

    def __repr__(self):
        return f"Type[{self.full_qualified_name()}]"

    def __str__(self):
        return f"Type[{self.full_qualified_name()}]"



from __future__ import annotations

from copy import deepcopy
from pprint import pprint
from typing import TYPE_CHECKING, Dict, Generator, List, NoReturn, Tuple

from aasm.preprocessor.preprocessor import Preprocessor
from aasm.utils.exception import PanicException


if TYPE_CHECKING:
    from aasm.intermediate.action import Action
    from aasm.intermediate.agent import Agent
    from aasm.intermediate.behaviour import Behaviour
    from aasm.intermediate.graph import Graph
    from aasm.intermediate.message import Message
    from aasm.intermediate.module import Module
    from aasm.modules.module import Module as LoadedModule


class ParsedData:
    def __init__(
        self, agents: List[Agent], messages: List[Message], graph: Graph | None
    ):
        self.agents: List[Agent] = agents
        self.messages: List[Message] = messages
        self.graph: Graph | None = graph


class State:
    def __init__(self, lines: List[str], modules: List[LoadedModule], debug: bool):
        self.debug: bool = debug
        self.preprocessor = Preprocessor(lines)
        self.lines: List[str] = self.preprocessor.run()
        self.line_num: int = 0
        self.in_agent: bool = False
        self.in_message: bool = False
        self.in_behaviour: bool = False
        self.in_action: bool = False
        self.in_graph: bool = False
        self.agents: Dict[str, Agent] = {}
        self.messages: Dict[Tuple[str, str], Message] = {}
        self.graph: Graph | None = None
        # TODO: Refactor this to a single module type
        self.modules: Dict[str, Module] = {}
        self.loaded_modules: List[LoadedModule] = modules

    @property
    def last_agent(self) -> Agent:
        return self.agents[list(self.agents.keys())[-1]]

    @property
    def last_behaviour(self) -> Behaviour:
        return self.last_agent.last_behaviour

    @property
    def last_action(self) -> Action:
        return self.last_agent.last_behaviour.last_action

    @property
    def last_message(self) -> Message:
        return self.messages[list(self.messages.keys())[-1]]

    @property
    def last_graph(self) -> Graph:
        if self.graph is None:
            raise Exception("Graph is not defined")
        return self.graph

    def add_agent(self, agent: Agent) -> None:
        self.agents[agent.name] = agent
        new_list_name = f"{agent.name}_list"
        for ag in self.agents:
            self.agents[ag].add_connection_list(new_list_name)

    def add_message(self, message: Message) -> None:
        self.messages[(message.type, message.performative)] = message

    def add_graph(self, graph: Graph) -> None:
        self.graph = graph

    def add_module(self, module: Module) -> None:
        self.modules[module.name] = module

    def agent_exists(self, name: str) -> bool:
        return name in self.agents

    def message_exists(self, msg_type: str, msg_performative: str) -> bool:
        return (msg_type, msg_performative) in self.messages

    def graph_exists(self) -> bool:
        return self.graph is not None

    def module_exists(self, name: str) -> bool:
        return name in self.modules

    def module_is_loaded(self, name: str) -> bool:
        return name in [module.name for module in self.loaded_modules]

    def get_message_instance(self, msg_type: str, msg_performative: str) -> Message:
        return deepcopy(self.messages[(msg_type, msg_performative)])

    def get_module_types(self) -> List[str]:
        ret_list = []
        for module in self.modules.values():
            loaded_module = self._find_module(module.name)
            if loaded_module is not None:
                ret_list += [mod_type.name for mod_type in loaded_module.types]
        return ret_list

    def get_init_function(self, type_name: str) -> str:
        for module in self.modules.values():
            loaded_module = self._find_module(module.name)
            if loaded_module is not None:
                for mod_type in loaded_module.types:
                    if mod_type.name == type_name:
                        return f"{module.name}.{mod_type.name}__init()"
        return ""

    def _find_module(self, name):
        for module in self.loaded_modules:
            if module.name == name:
                return module
        return None

    def tokens_from_lines(self) -> Generator[list[str], None, None]:
        for line in self.lines:
            self.line_num += 1
            uncommented = line.split("#")[0]
            tokens = [token.strip() for token in uncommented.replace(",", " ").split()]
            if tokens:
                tokens[0] = tokens[0].upper()
                yield tokens

    def print(self) -> None:
        if self.messages:
            print("- Messages:")
            for message in self.messages.values():
                message.print()
                print("")
        if self.agents:
            print("- Agents:")
            for agent in self.agents.values():
                agent.print()
                print("")
        if self.graph:
            print("- Graph:")
            self.graph.print()

    def verify_end_state(self) -> None:
        if self.in_agent:
            self.panic("Missing EAGENT")
        elif self.in_message:
            self.panic("Missing EMESSAGE")
        elif self.in_graph:
            self.panic("Missing EGRAPH")

    def get_parsed_data(self) -> ParsedData:
        self.verify_end_state()
        if self.debug:
            pprint(self.__dict__)
            self.print()
        return ParsedData(
            list(self.agents.values()), list(self.messages.values()), self.graph
        )

    def panic(self, reason: str, suggestion: str = "") -> NoReturn:
        if self.debug:
            pprint(self.__dict__)
            self.print()
        (err_line, err_data) = self.preprocessor.get_original_line_number(self.line_num)
        if err_data != "":
            place = f"Error in preprocessor directive: {err_data}, declared at line {err_line}"
        else:
            place = f"Error in line {err_line}: {self.lines[self.line_num - 1].strip()}"
        raise PanicException(place, reason, suggestion)

    def require(
        self, expr: bool, msg_on_error: str, suggestion_on_error: str = ""
    ) -> None:
        if not expr:
            self.panic(msg_on_error, suggestion_on_error)



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.intermediate.graph import (
    InhomogenousRandomGraph,
    InhomogeneousAgent,
    AgentConstantAmount,
    ConnectionConstantAmount,
)


from aasm.utils.validation import is_int, is_float

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_DEFTYPE(
    state: State, agent_name: str, amount: str, conn_amounts: List[str]
) -> None:
    state.require(state.agent_exists(agent_name), f"Agent {agent_name} is not defined.")
    state.require(
        not state.last_graph.is_agent_defined(agent_name),
        f"Agent {agent_name} is already defined.",
    )
    state.require(
        state.in_graph,
        "Cannot define agent type outside of graph scope.",
        "Try defining new graph using GRAPH.",
    )
    state.require(
        isinstance(state.last_graph, InhomogenousRandomGraph),
        "DEFTYPE can be used only in irg graphs.",
        "Define irg graphs with GRAPH irg.",
    )
    state.require(
        state.last_graph.is_types_amount_defined(),
        "The amount of types is not defined.",
        "Define the amount of types using TYPES.",
    )
    agent_amount: AgentConstantAmount | None = None
    state.require(is_int(amount), f"{amount} is not a valid integer.")
    agent_amount = AgentConstantAmount(amount)
    connection_amounts: List[ConnectionConstantAmount] = []
    state.require(
        len(conn_amounts) == state.last_graph.get_types_amount(),
        f"Amount of connection amounts ({len(conn_amounts)}) does not match the amount of types ({state.last_graph.get_types_amount()}).",
    )
    for conn_amount in conn_amounts:
        state.require(
            is_float(conn_amount), f"{conn_amount} is not a valid floating number."
        )
        state.require(
            float(conn_amount) >= 0,
            f"{conn_amount} is less than zero.",
            "Connection amounts must be between 0 and 100.",
        )
        state.require(
            float(conn_amount) <= 100,
            f"{conn_amount} is greater than 100.",
            "Connection amounts must be between 0 and 100.",
        )
        connection_amounts.append(ConnectionConstantAmount(conn_amount))

    state.last_graph.add_agent(
        InhomogeneousAgent(agent_name, agent_amount, connection_amounts)
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.utils.validation import is_int

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_MPARAMS(state: State, m0: str, m_inc: str) -> None:
    state.require(
        state.in_graph,
        "Not inside any graph",
        "MPARAMS can only be used inside graphs.",
    )
    state.require(not state.last_graph.is_m_defined(), "MPARAMS already defined.")
    state.require(is_int(m0) and is_int(m_inc), "MPARAMS must be integers.")
    m_begin = int(m0)
    m_increment = int(m_inc)
    state.require(m_begin >= 0 and m_increment >= 0, "MPARAMS must be non-negative.")

    state.last_graph.set_m((m_begin, m_increment))



import pytest
import os

import random
import numpy
import json

FIXTURE_PATH = os.path.join(os.path.dirname(__file__), "fixtures")


@pytest.fixture(autouse=True)
def set_random_seed():
    # Set a fixed seed value for numpy and random number generators.
    seed_value = 42
    numpy.random.seed(seed_value)
    random.seed(seed_value)


@pytest.fixture(scope="package")
def irg_graph_code():
    with open(os.path.join(FIXTURE_PATH, "inhomogenous.aasm")) as f:
        return f.read().split("\n")


@pytest.fixture(scope="package")
def statistical_graph_code():
    with open(os.path.join(FIXTURE_PATH, "statistical.aasm")) as f:
        return f.read().split("\n")


@pytest.fixture(scope="package")
def matrix_graph_code():
    with open(os.path.join(FIXTURE_PATH, "matrix.aasm")) as f:
        return f.read().split("\n")


@pytest.fixture(scope="package")
def barabasi_graph_code():
    with open(os.path.join(FIXTURE_PATH, "barabasi-albert.aasm")) as f:
        return f.read().split("\n")


@pytest.fixture(scope="package")
def algo_runner():
    def run_algorithm(graph_code_lines, domain, sim_id):
        exec("\n".join(graph_code_lines))
        try:
            algorithm = locals()["generate_graph_structure"]
        except KeyError:
            return []
        return algorithm(domain, sim_id)

    return run_algorithm



import pytest
import numpy
import random
import pprint
import re

from aasm import get_spade_code


def get_agent_names(code):
    agent_names = []
    for line in code:
        line = line.strip()
        tmp_line = line.upper()
        if (
            tmp_line.startswith("DEFG")
            or tmp_line.startswith("DEFNODE")
            or tmp_line.startswith("DEFTYPE")
        ):
            line = re.sub(r"\s+", " ", line)
            line = re.sub(r",", " ", line)
            agent_names.append(line.split(" ")[1].strip())
    return list(set(agent_names))


def assert_common_properties(generated_agents, expected_length, agent_names):
    assert (
        len(generated_agents) == expected_length
    ), f"Expected the graph to contain {expected_length} nodes, got {len(generated_agents)} instead"
    list_names = []
    for name in agent_names:
        list_name = name + "_list"
        list_names.append(list_name)
    for agent in generated_agents:
        assert (
            agent["type"] in agent_names
        ), f"Generated {agent['type']} not declared in code"
        sum_in_lists = 0
        print(len(agent["connections"]))
        print(agent["connections"])
        for list_name in list_names:
            assert (
                list_name in agent
            ), f"Generated {agent['type']} does not contain {list_name}"
            print(list_name)
            print(agent[list_name])
            print(len(agent[list_name]))
            sum_in_lists += len(agent[list_name])
            print(sum_in_lists)
        assert sum_in_lists == len(
            agent["connections"]
        ), f'agent {agent["jid"]} has {len(agent["connections"])} connections, but {sum_in_lists} in agent specific lists'


def test_irg_graph(irg_graph_code, algo_runner):
    spade_code = get_spade_code(irg_graph_code)
    agent_names = get_agent_names(irg_graph_code)
    generated = algo_runner(spade_code.graph_code_lines, "irg", "test_id")
    # pprint.pprint(generated)
    assert_common_properties(generated, 23, agent_names)
    null_count = 0
    for agent in generated:
        if agent["type"] == "null":
            assert (
                len(agent["null_list"]) == 0
            ), f'null agent {agent["jid"]} has non-zero null_list'
            assert (
                len(agent["media_source_list"]) == 0
            ), f'null agent {agent["jid"]} has non-zero media_source_list'
        if agent["type"] == "user":
            assert (
                len(agent["media_source_list"]) == 2
            ), f'user agent {agent["jid"]} has an incomplete media_source_list'
            if len(agent["null_list"]) == 0:
                null_count += 1
    # this test depends on the random seed
    assert (
        null_count == 6
    ), f"Expected 6 user agents, connected to null ones got {null_count}"


def test_statistical_graph(statistical_graph_code, algo_runner):
    spade_code = get_spade_code(statistical_graph_code)
    agent_names = get_agent_names(statistical_graph_code)
    generated = algo_runner(spade_code.graph_code_lines, "statistical", "test_id")
    pprint.pprint(generated)
    assert_common_properties(generated, 170, agent_names)
    # ensure 20 users and 150 media_source agents
    user_count = 0
    media_source_count = 0
    for agent in generated:
        conn_list_len = len(agent["connections"])
        sum_other_lists = len(agent["user_list"]) + len(agent["media_source_list"])
        assert (
            conn_list_len == sum_other_lists
        ), f'agent {agent["jid"]} has {conn_list_len} connections, but {sum_other_lists} in agent specific lists'
        if agent["type"] == "user":
            user_count += 1
            assert (
                conn_list_len == 10
            ), f'user agent {agent["jid"]} has {conn_list_len} connections expected: 10'
        if agent["type"] == "media_source":
            media_source_count += 1
            assert (
                conn_list_len > 0
            ), f'media_source agent {agent["jid"]} has 0 connections'
    assert user_count == 20, f"Expected 20 user agents, got {user_count}"
    assert (
        media_source_count == 150
    ), f"Expected 150 media_source agents, got {media_source_count}"


def test_matrix_graph(matrix_graph_code, algo_runner):
    spade_code = get_spade_code(matrix_graph_code)
    agent_names = get_agent_names(matrix_graph_code)
    generated = algo_runner(spade_code.graph_code_lines, "matrix", "test_id")
    pprint.pprint(generated)
    assert_common_properties(generated, 15, agent_names)
    scale = 3  # from fixture, could be moved to a parsing function from code
    user_count = 0
    media_source_count = 0
    null_count = 0
    for agent in generated:
        if agent["type"] == "user":
            user_count += 1
            user_list_len = len(agent["user_list"])
            assert (
                user_list_len >= 1 + scale - 1 and user_list_len <= 2 + scale - 1
            ), f'user agent {agent["jid"]} has {user_list_len} user agents, expected between {1 + scale - 1} and {2 + scale - 1}'
        elif agent["type"] == "media_source":
            media_source_count += 1
            assert (
                len(agent["connections"]) == 3 + scale - 1
            ), f'media_source agent {agent["jid"]} has {len(agent["connections"])} connections, expected {3 + scale - 1}'
            assert (
                len(agent["media_source_list"]) == scale - 1
            ), f'media_source agent {agent["jid"]} has {len(agent["media_source_list"])} media_source agents, expected {scale - 1}'
        elif agent["type"] == "null":
            null_count += 1
            assert (
                len(agent["connections"]) == 1 + scale - 1
            ), f'media_source agent {agent["jid"]} has {len(agent["connections"])} connections, expected {1 + scale - 1}'
            assert (
                len(agent["null_list"]) == scale - 1
            ), f'media_source agent {agent["jid"]} has {len(agent["null_list"])} null agents, expected {scale - 1}'
    assert (
        user_count == 3 * scale
    ), f"Expected {3 * scale} user agents, got {user_count}"
    assert (
        media_source_count == 1 * scale
    ), f"Expected {3 * scale} media_source agents, got {media_source_count}"
    assert (
        null_count == 1 * scale
    ), f"Expected {3 * scale} null agents, got {null_count}"


def test_barabasi_graph(barabasi_graph_code, algo_runner):
    spade_code = get_spade_code(barabasi_graph_code)
    agent_names = get_agent_names(barabasi_graph_code)
    generated = algo_runner(spade_code.graph_code_lines, "barabasi", "test_id")
    # pprint.pprint(generated)
    assert_common_properties(generated, 60, agent_names)
    user_count = 0
    media_source_count = 0
    for agent in generated:
        if agent["type"] == "user":
            user_count += 1
        elif agent["type"] == "media_source":
            media_source_count += 1
    assert user_count == 50, f"Expected 10 user agents, got {user_count}"
    assert (
        media_source_count == 10
    ), f"Expected 10 media_source agents, got {media_source_count}"



from __future__ import annotations

from itertools import compress
from typing import TYPE_CHECKING, List

from aasm.generating.python_code import PythonCode
from aasm.intermediate.graph import (
    AgentConstantAmount,
    AgentPercentAmount,
    ConnectionConstantAmount,
    ConnectionDistExpAmount,
    ConnectionDistNormalAmount,
    ConnectionDistUniformAmount,
    MatrixGraph,
    StatisticalGraph,
    BarabasiGraph,
    InhomogenousRandomGraph,
)

if TYPE_CHECKING:
    from aasm.intermediate.graph import Graph


class PythonGraph(PythonCode):
    def __init__(self, indent_size: int, graph: Graph | None):
        super().__init__(indent_size)
        if graph:
            self.add_required_imports()
            self.generate_graph(graph)

    def add_agent_class_definition(self):
        self.add_line("class Agent:")
        self.indent_right()
        self.add_line(
            "def __init__(self, jid: str, type: str, connections = None, sim_id: str = '', conn_list_names = None):"
        )
        self.indent_right()
        self.add_line("self.jid = jid")
        self.add_line("self.type = type")
        self.add_line("self.sim_id = sim_id")
        self.add_line("if conn_list_names is not None:")
        self.indent_right()
        self.add_line("self.conn_lists = {}")
        self.add_line("for name in conn_list_names:")
        self.indent_right()
        self.add_line(f"self.conn_lists[name + '_list'] = []")
        self.indent_left()
        self.indent_left()
        self.add_line("else:")
        self.indent_right()
        self.add_line("self.conn_lists = {}")
        self.indent_left()
        self.add_line("if connections is not None:")
        self.indent_right()
        self.add_line("self.connections = connections")
        self.indent_left()
        self.add_line("else:")
        self.indent_right()
        self.add_line("self.connections = []")
        self.indent_left()
        self.indent_left()
        self.add_newline()
        self.add_line("def add_connection(self, connection: str, type: str):")
        self.indent_right()
        self.add_line("self.connections.append(connection)")
        self.add_line("self.conn_lists[type + '_list'].append(connection)")
        self.indent_left()
        self.add_newline()
        self.add_line("def add_to_list(self, connection: str, type: str):")
        self.indent_right()
        self.add_line("self.conn_lists[type + '_list'].append(connection)")
        self.indent_left()
        self.add_newline()
        self.add_line("def to_dict(self):")
        self.indent_right()
        self.add_line("ret = {}")
        self.add_line("ret['jid'] = self.jid")
        self.add_line("ret['type'] = self.type")
        self.add_line("ret['connections'] = self.connections")
        self.add_line("ret['sim_id'] = self.sim_id")
        self.add_line("for name, conn_list in self.conn_lists.items():")
        self.indent_right()
        self.add_line("ret[name] = conn_list")
        self.indent_left()
        self.add_line("return ret")
        self.indent_left()
        self.indent_left()

    def add_required_imports(self) -> None:
        self.add_line("import random")
        self.add_line("import uuid")
        self.add_line("import numpy")

    def generate_graph(self, graph: Graph) -> None:
        match graph:
            case StatisticalGraph():
                self.add_statistical_graph(graph)

            case MatrixGraph():
                self.add_matrix_graph(graph)

            case BarabasiGraph():
                self.add_barabasi_graph(graph)

            case InhomogenousRandomGraph():
                self.add_irg_graph(graph)

            case _:
                raise Exception(f"Unknown graph type: {graph.print()}")

    def add_statistical_graph(self, graph: StatisticalGraph) -> None:
        self.add_line('def generate_graph_structure(domain, sim_id=""):')
        self.indent_right()
        self.add_agent_class_definition()

        if not graph.agents:
            self.add_line("return []")
            self.indent_left()
            return
        num_agents_expr: List[str] = []
        self.add_line("agent_types = []")
        for agent in graph.agents.values():
            self.add_line(f"agent_types += ['{agent.name}']")
            match agent.amount:
                case AgentConstantAmount():
                    self.add_line(f"_num_{agent.name} = {agent.amount.value}")

                case AgentPercentAmount():
                    self.add_line(
                        f"_num_{agent.name} = round({agent.amount.value} / 100 * {graph.size})"
                    )

                case _:
                    raise Exception(
                        f"Unknown agent amount type: {agent.amount.print()}"
                    )

            num_agents_expr.append(f"_num_{agent.name}")

        self.add_line(f'num_agents = {" + ".join(num_agents_expr)}')

        self.add_line('if sim_id == "":')
        self.indent_right()
        self.add_line("random_id = str(uuid.uuid4())[:5]")
        self.indent_left()
        self.add_line("else:")
        self.indent_right()
        self.add_line("random_id = sim_id")
        self.indent_left()
        self.add_line('jids = [f"{i}_{random_id}@{domain}" for i in range(num_agents)]')
        self.add_line("agents = []")
        self.add_line("next_agent_idx = 0")
        for agent in graph.agents.values():
            self.add_line(f"for _ in range(_num_{agent.name}):")
            self.indent_right()

            match agent.connections:
                case ConnectionConstantAmount():
                    self.add_line(f"num_connections = {agent.connections.value}")

                case ConnectionDistNormalAmount():
                    self.add_line(
                        f"num_connections = int(numpy.random.normal({agent.connections.mean}, {agent.connections.std_dev}))"
                    )

                case ConnectionDistExpAmount():
                    self.add_line(
                        f"num_connections = int(numpy.random.exponential(1 / {agent.connections.lambda_}))"
                    )

                case ConnectionDistUniformAmount():
                    self.add_line(
                        f"num_connections = int(random.uniform({agent.connections.a}, {agent.connections.b}))"
                    )

                case _:
                    raise Exception(
                        f"Unknown connection amount: {agent.connections.print()}"
                    )

            self.add_line(
                "num_connections = max(min(num_connections, len(jids) - 1), 0)"
            )
            self.add_line("jid = jids[next_agent_idx]")
            self.add_line(
                "connections = random.sample([other_jid for other_jid in jids if other_jid != jid], num_connections)"
            )
            self.add_line(
                f"new_agent = Agent(jid, '{agent.name}', connections, random_id, agent_types)"
            )
            self.add_line("agents.append(new_agent)")
            self.add_line("next_agent_idx += 1")
            self.indent_left()
        self.add_line("for agent in agents:")
        self.indent_right()
        self.add_line("for other_agent in agents:")
        self.indent_right()
        self.add_line("if agent.jid in other_agent.connections:")
        self.indent_right()
        self.add_line("other_agent.add_to_list(agent.jid, agent.type)")
        self.indent_left()
        self.indent_left()
        self.indent_left()
        self.add_line("return [agent.to_dict() for agent in agents]")
        self.indent_left()

    def add_matrix_graph(self, graph: MatrixGraph):
        self.add_line('def generate_graph_structure(domain, sim_id=""):')
        self.indent_right()
        self.add_agent_class_definition()
        if not graph.agents:
            self.add_line("return []")
            self.indent_left()
            return
        if graph.is_scale_defined():
            self.add_line(f"scale_factor = {graph.scale}")
        else:
            self.add_line(f"scale_factor = 1")
        self.add_line(f"n_agent_types = {len(graph.agents)}")
        self.add_line("graph_size = scale_factor * n_agent_types")
        self.add_line('if sim_id == "":')
        self.indent_right()
        self.add_line("random_id = str(uuid.uuid4())[:5]")
        self.indent_left()
        self.add_line("else:")
        self.indent_right()
        self.add_line("random_id = sim_id")
        self.indent_left()
        self.add_line('jids = [f"{i}_{random_id}@{domain}" for i in range(graph_size)]')
        self.add_line("agent_types = []")
        self.add_line("agents = []")
        self.add_line("indx_sets = []")
        for agent in graph.agents:
            self.add_line(f"agent_types.append('{agent.name}')")
            adj_indx = [True if x == 1 else False for x in agent.adj_row.row]
            adj_indx = list(compress(range(len(adj_indx)), adj_indx))
            self.add_line(f"indx_sets.append({adj_indx})")

        self.add_line("agent_types = agent_types * scale_factor")

        self.add_line("for base_agent_index in range(n_agent_types):")
        self.indent_right()
        self.add_line("indices = indx_sets[base_agent_index]")
        self.add_line("for shift in range(1, scale_factor):")
        self.indent_right()
        self.add_line(
            f"indices.append((base_agent_index + shift * n_agent_types) % graph_size)"
        )
        self.indent_left()
        self.add_line("for shift in range(scale_factor):")
        self.indent_right()
        self.add_line("agent_idx = base_agent_index + shift * n_agent_types")
        self.add_line("jid = jids[agent_idx]")
        self.add_line("agent_type = agent_types[agent_idx]")
        self.add_line("connections = []")
        self.add_line("for i in indices:")
        self.indent_right()
        self.add_line(
            f"connections.append(jids[(i + shift * n_agent_types) % graph_size])"
        )
        self.indent_left()
        self.add_line(
            "new_agent = Agent(jid, agent_type, connections, random_id, list(set(agent_types)))"
        )
        self.add_line("agents.append(new_agent)")
        self.indent_left()
        self.indent_left()
        self.add_line("for agent in agents:")
        self.indent_right()
        self.add_line("for other_agent in agents:")
        self.indent_right()
        self.add_line("if agent.jid in other_agent.connections:")
        self.indent_right()
        self.add_line("other_agent.add_to_list(agent.jid, agent.type)")
        self.indent_left()
        self.indent_left()
        self.indent_left()
        self.add_line("return [agent.to_dict() for agent in agents]")
        self.indent_left()

    def add_barabasi_graph(self, graph: BarabasiGraph):
        # generate jid list using draw by draw Barabasi-Albert algorithm for the provided graph with its m0 and m
        self.add_line('def generate_graph_structure(domain, sim_id=""):')
        self.indent_right()
        self.add_agent_class_definition()
        if not graph.agents:
            self.add_line("return []")
            self.indent_left()
            return

        self.add_line(f"agent_types = []")

        num_agents_expr: List[str] = []
        for agent in graph.agents.values():
            match agent.amount:
                case AgentConstantAmount():
                    self.add_line(f"_num_{agent.name} = {agent.amount.value}")

                case AgentPercentAmount():
                    self.add_line(
                        f"_num_{agent.name} = round({agent.amount.value} / 100 * {graph.size})"
                    )

                case _:
                    raise Exception(
                        f"Unknown agent amount type: {agent.amount.print()}"
                    )

            num_agents_expr.append(f"_num_{agent.name}")
            self.add_line(f"tmp = ['{agent.name}'] * _num_{agent.name}")
            self.add_line("agent_types.extend(tmp)")

        self.add_line(f'num_agents = {" + ".join(num_agents_expr)}')
        self.add_line("random.shuffle(agent_types)")

        self.add_line('if sim_id == "":')
        self.indent_right()
        self.add_line("random_id = str(uuid.uuid4())[:5]")
        self.indent_left()
        self.add_line("else:")
        self.indent_right()
        self.add_line("random_id = sim_id")
        self.indent_left()
        if graph.m_params is not None:
            self.add_line(f"m0 = {graph.m_params.m0}")
            self.add_line(f"m = {graph.m_params.m_inc}")
        self.add_line('jids = [f"{i}_{random_id}@{domain}" for i in range(num_agents)]')
        # self.add_line("agents = []")
        self.add_line("connection_lists = []")
        self.add_line("next_agent_idx = 0")
        # begin by creating the initial complete graph
        self.add_line("init_jids = random.choices(jids, k=m0)")
        self.add_line("for jid in init_jids:")
        self.indent_right()
        self.add_line("copy = init_jids.copy()")
        self.add_line("copy.remove(jid)")
        self.add_line("connection_lists.append(copy)")
        self.indent_left()
        self.add_line("next_agent_idx += m0")
        self.add_line("for i in range(m0, num_agents):")
        self.indent_right()
        self.add_line("to_connect = []")
        self.add_line("ids = list(range(next_agent_idx))")
        self.add_line("while len(to_connect) < m:")
        self.indent_right()
        self.add_line("weights = [len(connection_lists[ident]) for ident in ids]")
        self.add_line("to_connect.append(random.choices(ids, weights=weights)[0])")
        self.add_line("ids.remove(to_connect[-1])")
        self.indent_left()
        self.add_line("to_connect_jids = [jids[ident] for ident in to_connect]")
        self.add_line("connection_lists.append(to_connect_jids)")
        self.add_line("for id in to_connect:")
        self.indent_right()
        # two sided connections
        self.add_line("connection_lists[id].append(jids[next_agent_idx])")
        self.add_line("connection_lists[id] = list(set(connection_lists[id]))")
        self.indent_left()
        self.add_line("next_agent_idx += 1")
        self.indent_left()
        self.add_line("agents = []")
        self.add_line("for i in range(num_agents):")
        self.indent_right()
        self.add_line(
            "agent = Agent(jids[i], agent_types[i], connection_lists[i], random_id, list(set(agent_types)))"
        )
        self.add_line("agents.append(agent)")
        self.indent_left()
        self.add_line("for agent in agents:")
        self.indent_right()
        self.add_line("for other_agent in agents:")
        self.indent_right()
        self.add_line("if agent.jid in other_agent.connections:")
        self.indent_right()
        self.add_line("other_agent.add_to_list(agent.jid, agent.type)")
        self.indent_left()
        self.indent_left()
        self.indent_left()
        self.add_line("return [agent.to_dict() for agent in agents]")
        self.indent_left()

    def add_irg_graph(self, graph: InhomogenousRandomGraph):
        self.add_line('def generate_graph_structure(domain, sim_id=""):')
        self.indent_right()
        self.add_newline()
        self.add_agent_class_definition()
        self.add_newline()
        if not graph.agents:
            self.add_line("return []")
            self.indent_left()
            return
        self.add_line("agent_types = []")
        self.add_line("probo_matrix = {}")
        num_agents_expr: List[str] = []
        for agent in graph.agents.values():
            self.add_line(f"probo_matrix['{agent.name}'] = []")
            for probo in agent.amounts:
                self.add_line(f"probo_matrix['{agent.name}'].append({probo.value})")
            match agent.amount:
                case AgentConstantAmount():
                    self.add_line(f"_num_{agent.name} = {agent.amount.value}")

                case AgentPercentAmount():
                    self.add_line(
                        f"_num_{agent.name} = round({agent.amount.value} / 100 * {graph.size})"
                    )

                case _:
                    raise Exception(
                        f"Unknown agent amount type: {agent.amount.print()}"
                    )

            num_agents_expr.append(f"_num_{agent.name}")
            self.add_line(f"tmp = ['{agent.name}'] * _num_{agent.name}")
            self.add_line("agent_types.extend(tmp)")

        self.add_line(f'num_agents = {" + ".join(num_agents_expr)}')

        self.add_line('if sim_id == "":')
        self.indent_right()
        self.add_line("random_id = str(uuid.uuid4())[:5]")
        self.indent_left()
        self.add_line("else:")
        self.indent_right()
        self.add_line("random_id = sim_id")
        self.indent_left()
        self.add_line('jids = [f"{i}_{random_id}@{domain}" for i in range(num_agents)]')
        self.add_line(f"order = {graph.order}")
        self.add_line("agents = []")
        self.add_line("for i in range(num_agents):")
        self.indent_right()
        self.add_line(
            "agents.append(Agent(jids[i], agent_types[i], sim_id = random_id, conn_list_names=agent_types))"
        )
        self.indent_left()
        self.add_line("for agent in agents:")
        self.indent_right()
        self.add_line("for agent2 in agents:")
        self.indent_right()
        self.add_line("if agent != agent2:")
        self.indent_right()
        self.add_line("conn_index = order.index(agent2.type)")
        self.add_line("probo = probo_matrix[agent.type][conn_index]")
        self.add_line("if random.random() * 100 < probo:")
        self.indent_right()
        self.add_line("agent.add_connection(agent2.jid, agent2.type)")
        self.indent_left()
        self.indent_left()
        self.indent_left()
        self.indent_left()
        self.add_line("return [agent.to_dict() for agent in agents]")
        self.indent_left()



from __future__ import annotations

from typing import List, Set


class PythonCode:
    def __init__(self, indent_size: int):
        self.indent_size = indent_size
        self.indent: int = 0
        self.code_lines: List[str] = []
        self.required_imports: Set[str] = set()

    def indent_left(self) -> None:
        self.indent -= self.indent_size

    def indent_right(self) -> None:
        self.indent += self.indent_size

    def add_line(
        self, line: str, required_imports: Set[str] | None = None, add_newline=True
    ) -> None:
        if required_imports is not None:
            self.required_imports.update(required_imports)
        if add_newline:
            self.code_lines.append(self.indent * " " + line + "\n")
        else:
            self.code_lines.append(self.indent * " " + line)

    def add_newline(self) -> None:
        self.add_line("")

    def add_newlines(self, count: int) -> None:
        for _ in range(count):
            self.add_newline()

    def add_required_imports(self) -> None:
        lines: List[str] = []
        for required_import in self.required_imports:
            lines.append(f"import {required_import}\n")
        lines.sort()
        self.code_lines = lines + self.code_lines



from __future__ import annotations

from typing import TYPE_CHECKING, List as TypingList, Dict
from aasm.intermediate.argument import Argument
from aasm.modules.type import Type

from aasm.utils.exception import PanicException

from aasm.intermediate.argument import (
    Mutable,
    Float,
    Enum,
    EnumValue,
    ConnectionList,
    MessageList,
    FloatList,
    List,
    Connection,
    Literal,
    Message,
)
from aasm.intermediate.instruction import ModuleInstruction

if TYPE_CHECKING:
    from aasm.parsing.state import State

from pprint import pprint


class Instruction:
    def __init__(
        self,
        module_name: str,
        available_types: TypingList[Type],
        opcode: str,
        args: TypingList[str],
    ):
        self.module = module_name
        if opcode.endswith("*"):
            self.opcode = opcode[:-1].upper()
            self.is_block = True
        else:
            self.opcode = opcode.upper()
            self.is_block = False
        self.available_types = [
            Type("mut", "base"),
            Type("float", "base"),
            Type("enum", "base"),
            Type("enum_val", "base"),
            Type("list_conn", "base"),
            Type("list_msg", "base"),
            Type("list_float", "base"),
            Type("list", "base"),
            Type("conn", "base"),
            Type("literal", "base"),
            Type("message", "base"),
        ]
        self.available_types.extend(available_types)
        self.args_dict: Dict[str, TypingList[Type]] = {}
        self.arg_names = []
        self.assignment = None
        self._parse_args(args)
        self._validate_args_dict()
        self._set_assignment()

    def _set_assignment(self):
        # find the mutable value in args_dict
        mutable_arg = None
        for arg_name in self.args_dict.keys():
            for arg_type in self.args_dict[arg_name]:
                if arg_type == Type("mut", "base"):
                    mutable_arg = arg_name
        self.assignment = mutable_arg

    def _validate_args_dict(self):
        # ensure that at most one argument is mutable
        mutable_count = 0
        for arg_name in self.args_dict.keys():
            for arg_type in self.args_dict[arg_name]:
                if arg_type == Type("mut", "base"):
                    mutable_count += 1
        if mutable_count > 1:
            raise PanicException(
                f"Error in module {self.module}, instruction {self.opcode}.",
                f"Instruction {self.opcode} has more than one mutable argument.",
                "Ensure that at most one argument is mutable.",
            )
        elif mutable_count == 1 and self.is_block:
            raise PanicException(
                f"Error in module {self.module}, instruction {self.opcode}.",
                f"Instruction {self.opcode} is a block instruction and cannot have a mutable argument.",
                "Remove the mutable argument.",
            )

    def _parse_args(self, args: TypingList[str]):
        current_var_name = ""
        current_var_type = ""
        first_arg = True
        for arg in args:
            if arg.endswith(":"):
                if not first_arg:
                    self._validate_var_declaration(current_var_name)
                current_var_name = arg[:-1]
                self.arg_names.append(current_var_name)
                current_var_type = ""
                # verify that the variable name is not already used
                if current_var_name in self.args_dict:
                    raise PanicException(
                        f"Error in module {self.module}, instruction {self.opcode}. Variable {current_var_name} already defined.",
                        f"Variable {current_var_name} is already defined.",
                        "Rename the variable.",
                    )
                self.args_dict[current_var_name] = []
                first_arg = False
            else:
                current_var_type = self._get_arg_type_from_name(arg)
                if len(self.args_dict[current_var_name]) > 0:
                    if self.args_dict[current_var_name][0] != Type("mut", "base"):
                        raise PanicException(
                            f"Error in module {self.module}, instruction {self.opcode}. Variable {current_var_name} cannot have multiple types.",
                            f"Variable {current_var_name} cannot have multiple types.",
                            "Specify only one type for the variable.",
                        )
                self.args_dict[current_var_name].append(current_var_type)

    def _get_arg_type_from_name(self, arg_name: str) -> Type:
        for arg_type in self.available_types:
            if arg_type.name == arg_name:
                return arg_type
        raise PanicException(
            f"Error in module {self.module}, instruction {self.opcode}. Type {arg_name} is not defined.",
            f"Type {arg_name} is not defined.",
            "Define the type.",
        )

    def _validate_var_declaration(self, current_var_name: str):
        if len(self.args_dict[current_var_name]) == 0:
            raise PanicException(
                f"Error in module {self.module}, instruction {self.opcode}. Missing type for variable {current_var_name}.",
                f"Variable {current_var_name} has no type specified.",
                "Specify a type for the variable.",
            )

        if len(self.args_dict[current_var_name]) > 1:
            if Type("mut", "base") not in self.args_dict[current_var_name]:
                raise PanicException(
                    f"Error in module {self.module}, instruction {self.opcode}. Variable {current_var_name} cannot have multiple types.",
                    f"Variable {current_var_name} cannot have multiple types.",
                    "Specify only one type for the variable.",
                )
        # verify that the types associated with current_var_name are valid
        for var_type in self.args_dict[current_var_name]:
            if var_type not in self.available_types:
                raise PanicException(
                    f"Error in module {self.module}, instruction {self.opcode}. Type {var_type} is not defined.",
                    f"Type {var_type} is not defined.",
                    "Define the type.",
                )

    def op(self, state: State, arguments: TypingList[str]) -> None:
        state.require(
            state.in_action,
            "Not inside any action.",
            f"{self.opcode} can be used inside actions.",
        )
        state.require(
            len(arguments) == len(self.args_dict.keys()),
            f"Wrong number of arguments for {self.opcode}.",
            f"Expected {len(self.args_dict.keys())}, got {len(arguments)}.",
        )

        parsed_args = [Argument(state, arg) for arg in arguments]
        state.require(
            self._validate_types_in_op_context(parsed_args),
            f"Mismatched types in the {self.module}::{self.opcode} context.: {[arg.explain() for arg in parsed_args]}",
            f"Refer to module documentation for further help.",
        )

        instruction_assignment = None

        if self.assignment is not None:
            idx = 0
            for name in self.arg_names:
                if name == self.assignment:
                    instruction_assignment = parsed_args[idx]
                idx += 1

        state.last_action.add_instruction(
            ModuleInstruction(
                parsed_args,
                self.opcode,
                self.module,
                self.is_block,
                instruction_assignment,
            )
        )
        if self.is_block:
            state.last_action.start_block()

    def _validate_types_in_op_context(self, parsed_args) -> bool:
        arg_idx = 0
        for arg in parsed_args:
            types_to_check = []
            for arg_type in self.args_dict[self.arg_names[arg_idx]]:
                if arg_type == Type("mut", "base"):
                    types_to_check.append(Mutable)
                elif arg_type == Type("float", "base"):
                    types_to_check.append(Float)
                elif arg_type == Type("enum", "base"):
                    types_to_check.append(Enum)
                elif arg_type == Type("enum_val", "base"):
                    types_to_check.append(EnumValue)
                elif arg_type == Type("list_conn", "base"):
                    types_to_check.append(List)
                    types_to_check.append(ConnectionList)
                elif arg_type == Type("list_msg", "base"):
                    types_to_check.append(List)
                    types_to_check.append(MessageList)
                elif arg_type == Type("list_float", "base"):
                    types_to_check.append(List)
                    types_to_check.append(FloatList)
                elif arg_type == Type("list", "base"):
                    types_to_check.append(List)
                elif arg_type == Type("conn", "base"):
                    types_to_check.append(Connection)
                elif arg_type == Type("literal", "base"):
                    types_to_check.append(Literal)
                elif arg_type == Type("message", "base"):
                    types_to_check.append(Message)
                else:
                    new_type = arg.get_modvar_type(arg_type.name)
                    types_to_check.append(new_type)

            for arg_type in types_to_check:
                if not arg.has_type(arg_type):
                    return False
            arg.set_op_type(*types_to_check)
            arg_idx += 1

        return True

    def __str__(self) -> str:
        ret = f"{self.module}.{self.opcode}({', '.join(self.args_dict.keys())})"
        if self.is_block:
            ret += "[BLOCK]"
        return ret

    def __repr__(self) -> str:
        return str(self)



from __future__ import annotations

from typing import TYPE_CHECKING

from aasm.utils.validation import is_int

from aasm.intermediate.graph import InhomogenousRandomGraph

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_TYPES(state: State, types: str) -> None:
    state.require(
        state.in_graph, "Not inside any graph.", "SIZE can be used inside graphs."
    )
    state.require(
        isinstance(state.last_graph, InhomogenousRandomGraph),
        "TYPES can be used only in irg graphs.",
        "Define irg graphs with GRAPH irg.",
    )
    state.require(
        not state.last_graph.is_types_amount_defined(), "Types are already defined."
    )
    state.require(is_int(types), "Types must be an integer value.", "Example: TYPES 3")
    state.require(int(types) >= 0, "Types cannot be a negative number.")
    state.last_graph.set_types_amount(int(types))



from __future__ import annotations

from typing import TYPE_CHECKING
from aasm.intermediate.module import Module
from aasm.utils.validation import is_valid_name, print_invalid_names

if TYPE_CHECKING:
    from aasm.parsing.state import State


def op_MODULE(state: State, name: str) -> None:
    state.require(
        not state.in_agent,
        "Cannot define modules inside agents.",
        "First end current agent using EAGENT.",
    )
    state.require(
        not state.in_graph,
        "Cannot define modules inside graphs.",
        "First end current graph using EGRAPH.",
    )
    state.require(
        not state.module_exists(name),
        f"Module {name} already exists in the current environment.",
    )

    state.require(
        state.module_is_loaded(name),
        f"Module {name} is not loaded.",
    )

    # NOTE: I don't think that's necessary for modules
    # state.require(
    #     is_valid_name(name),
    #     f"{name} is not a correct name.",
    #     f"Names can only contain alphanumeric characters, underscores and cannot be: {print_invalid_names()}.",
    # )
    state.add_module(Module(name))



from __future__ import annotations

from typing import TYPE_CHECKING, List

from aasm.generating.python_code import PythonCode
from aasm.modules.module import Module, Type


def get_modules_for_target(
    module_code_lines: List[List[str]], target: str
) -> List[Module]:
    """
    Get all modules from `module_code_lines` that target the given `target`.

    :param module_code_lines: the code lines of all modules
    :param target: the target to filter by

    :return: the modules that target the given `target`
    """
    target_modules = []
    for module_lines in module_code_lines:
        module = Module(module_lines)
        print("Checking: ", module.name)
        if module.does_target(target):
            print("Adding: ", module.name)
            target_modules.append(module)
    return target_modules


class PythonModule(PythonCode):
    def __init__(self, indent_size, module: Module):
        super().__init__(indent_size)
        self.module: Module = module
        self.target: str = "spade"
        self.add_newlines(2)
        self.create_module_header()
        self.add_newline()
        self.create_module_imports()
        self.add_newline()
        self.create_module_implementations()

    def create_module_header(self) -> None:
        self.add_line(f"# Module: {self.module.name}")
        for line in self.module.description:
            self.add_line(f"# {line}", add_newline=False)

    def create_module_imports(self):
        try:
            preamble = self.module.preambles[self.target]
            for line in preamble:
                self.add_line(line)
        except KeyError:
            return

    def create_module_implementations(self):
        for type in self.module.types:
            impl_lines = type.init_lines[self.target]
            self.add_line(f"def {type.name}__init():")
            self.indent_right()
            for line in impl_lines:
                self.add_line(line)
            self.indent_left()
            self.add_newline()

        for impl in self.module.impls:
            if impl[0] == self.target:
                args_data = self.module.get_args_for_instruction(impl[1])
                args_string = ", ".join(args_data.keys())
                self.add_line(f"def {impl[1]}({args_string}):")
                self.indent_right()
                for line in self.module.impls[impl]:
                    self.add_line(line)
                self.indent_left()
                self.add_newline()
